// TODO: Put public facing types in this file.

// import 'dart:async';
// import 'dart:io';

import 'package:kenken/kenken.dart';

/// Checks if you are awesome. Spoiler: you are.
class Kenken {
  Kenken(this.size) : cells = List.generate(size, (_)=>List.filled(size, 0));

  int size;
  var cells = <List<int>>[];
  List<Rule> rules;

  /// Use it for set cell `value` by cell coordinates `x` and `y`.
  /// Coordinates should be in `1...kenken.size` diapazone.
  void setCell(int row, int col, int value) {
    var errMsg = _coordErrorMessage(row, col);
    if (errMsg.isNotEmpty) {
      throw ArgumentError(errMsg);
    }
    (cells[row-1])[col-1] = value;
  }

  /// Returns list of coordinates for row with given `rowNumber`
  List<RC> coordinatesOfRow(int rowNumber) =>
      List.generate(size, (i) => RC(rowNumber, i + 1));

  /// Returns list of coordinates for column with given `colNumber`
  List<RC> coordinatesOfCol(int colNumber) =>
      List.generate(size, (i) => RC(i + 1,colNumber));

  String _coordErrorMessage(int x, int y) {
    if (x < 1 || x > size || y < 1 || y > size) {
      return 'Bad coordinates: $x,$y';
    }
    return '';
  }

  /// Returns cell value
  int valueOf(RC cellCoord) {
    var errMsg = _coordErrorMessage(cellCoord.row, cellCoord.col);
    if (errMsg.isNotEmpty) {
      throw ArgumentError(errMsg);
    }
    return cells[cellCoord.row-1][cellCoord.col-1];
  }

  /// Returns invalid rules for kenken.
  /// If invalidRules list is empty, then kenken are solved.
  List<Rule> get invalidRules => rules.where((rule) {
        var ruleCellsValues =
            rule.cellsCoordinates.map((coords) => valueOf(coords)).toList();
        print('coordinates: ${rule.cellsCoordinates} values:$ruleCellsValues');
        var actualResult = rule.operation.predicate(ruleCellsValues);
        return actualResult.value != rule.validOperationResult.value;
      }).toList();
}

/// A rule validates list of cells.
/// List of rules validates a whole kenken.
class Rule {
  Rule(this.cellsCoordinates, this.operation, this.validOperationResult);

  List<RC> cellsCoordinates;
  Operation operation;
  OperationResult validOperationResult;
  @override
  String toString() =>
      '$operation on $cellsCoordinates should be $validOperationResult';
}

/// Cell coordinates. Coordinates should be in 1...kenken.size diapazone.
class RC {
  RC(this.row, this.col);
  int row;
  int col;
  @override
  String toString() => '($row,$col)';
}

class OperationResult<T> {
  OperationResult(this.value);
  T value;
  @override
  String toString() => '$value';
}

/// Operation will be apply to list of cells for validate a rule.
/// Operation has predicate (function to be apply)
/// and printable operation symbol, accesible by `toString()` method (for display it on kenken).
class Operation {
  Operation(this.symbol, this.predicate);

  String symbol;
  Predicate predicate;

  @override
  String toString() => symbol;
}

/// Function which can be applyed to List of values
typedef Predicate = OperationResult Function(List<int>);

/// Is list of values unique?
OperationResult isUniq(List<int> values) => OperationResult(values.isUniq);

extension ListUniq<T> on List<T> {
  /// Is List values uniq?
  bool get isUniq {
    if (length < 2) return true;
    for (var i = 0; i < length - 1; i++) {
      if (sublist(i + 1).contains(this[i])) return false;
    }
    return true;
  }
}
